#ifndef MMTRACE
#include <stdio.h>
#endif

#include "shader_common.h"
#include "nv40_shader.h"

extern ShaderSwzComp NV20VP_TX_SWIZZLE[4];
extern ShaderRegSwz  NV20VPTXSwizzle(int hwswz);

static int
NV40VPHasMergedInst(ShaderFunc *shader)
{
	if (shader->GetOpcodeHW(shader, 0) != NV40_VP_INST_OP_NOP &&
		shader->GetOpcodeHW(shader, 1) != NV40_VP_INST_OP_NOP)
		return 1;
	return 0;
}

static unsigned int
NV40VPGetOpcodeHW(ShaderFunc *shader, int slot)
{
	int op;

	if (slot)
		op = (shader->inst[1] & NV40_VP_INST_SCA_OPCODE_MASK)
			>> NV40_VP_INST_SCA_OPCODE_SHIFT;
	else
		op = (shader->inst[1] & NV40_VP_INST_VEC_OPCODE_MASK)
			>> NV40_VP_INST_VEC_OPCODE_SHIFT;

	return op;
}

static ShaderRegFile
NV40VPGetDestFile(ShaderFunc *shader, int merged)
{
	ShaderOpcode op;

	op = shader->GetOpcode(shader, merged);
	switch(op) {
	case SOP_ARL:
	case SOP_ARR:
	case SOP_ARA:
	case SOP_POPA:
		return SRF_ADDRESS;
	default:
		if (shader->GetOpcodeSlot(shader, merged)) {
			if (shader->inst[3] & NV40_VP_INST_SCA_RESULT)
				return SRF_RESULT;
		} else {
			if (shader->inst[0] & NV40_VP_INST_VEC_RESULT)
				return SRF_RESULT;
		}
		return SRF_TEMP;
	}

}

static unsigned int
NV40VPGetDestID(ShaderFunc *shader, int merged)
{
	int id;

	switch (shader->GetDestFile(shader, merged)) {
	case SRF_RESULT:
		id = ((shader->inst[3] & NV40_VP_INST_DEST_MASK)
				>> NV40_VP_INST_DEST_SHIFT);
		switch (id) {
		case NV40_VP_INST_DEST_POS  : return SR_POSITION;
		case NV40_VP_INST_DEST_COL0 : return SR_COL0;
		case NV40_VP_INST_DEST_COL1 : return SR_COL1;
		case NV40_VP_INST_DEST_BFC0 : return SR_BFC0;
		case NV40_VP_INST_DEST_BFC1 : return SR_BFC1;
		case NV40_VP_INST_DEST_FOGC :
		{ 
			int mask = shader->GetDestMask(shader, merged);
			switch (mask) {
			case SMASK_X: return SR_FOGCOORD;
			case SMASK_Y: return SR_CLIP0;
			case SMASK_Z: return SR_CLIP1;
			case SMASK_W: return SR_CLIP2;
			default:
				printf("more than 1 mask component set in FOGC writemask!\n");
				return SR_UNKNOWN;
			}
		}
		case NV40_VP_INST_DEST_PSZ  :
		{ 
			int mask = shader->GetDestMask(shader, merged);
			switch (mask) {
			case SMASK_X: return SR_POINTSZ;
			case SMASK_Y: return SR_CLIP3;
			case SMASK_Z: return SR_CLIP4;
			case SMASK_W: return SR_CLIP5;
			default:
				printf("more than 1 mask component set in PSZ writemask!\n");
				return SR_UNKNOWN;
			}
		}
		case NV40_VP_INST_DEST_TC(0): return SR_TEXCOORD0;
		case NV40_VP_INST_DEST_TC(1): return SR_TEXCOORD1;
		case NV40_VP_INST_DEST_TC(2): return SR_TEXCOORD2;
		case NV40_VP_INST_DEST_TC(3): return SR_TEXCOORD3;
		case NV40_VP_INST_DEST_TC(4): return SR_TEXCOORD4;
		case NV40_VP_INST_DEST_TC(5): return SR_TEXCOORD5;
		case NV40_VP_INST_DEST_TC(6): return SR_TEXCOORD6;
		case NV40_VP_INST_DEST_TC(7): return SR_TEXCOORD7;
		default:
			return -1;
		}
	case SRF_ADDRESS:
		/* Instructions that write address regs are encoded as if
		 * they would write temps.
		 */
	case SRF_TEMP:
		if (shader->GetOpcodeSlot(shader, merged))
			id = ((shader->inst[3] & NV40_VP_INST_SCA_DEST_TEMP_MASK)
					>> NV40_VP_INST_SCA_DEST_TEMP_SHIFT);
		else
			id = ((shader->inst[0] & NV40_VP_INST_VEC_DEST_TEMP_MASK)
					>> NV40_VP_INST_VEC_DEST_TEMP_SHIFT);
		return id;
	default:
		return -1;
	}
}

static unsigned int
NV40VPGetDestMask(ShaderFunc *shader, int merged)
{
	unsigned int mask = 0;

	if (shader->GetOpcodeSlot(shader, merged)) {
		if (shader->inst[3] & NV40_VP_INST_SCA_WRITEMASK_X) mask |= SMASK_X;
		if (shader->inst[3] & NV40_VP_INST_SCA_WRITEMASK_Y) mask |= SMASK_Y;
		if (shader->inst[3] & NV40_VP_INST_SCA_WRITEMASK_Z) mask |= SMASK_Z;
		if (shader->inst[3] & NV40_VP_INST_SCA_WRITEMASK_W) mask |= SMASK_W;
	} else {
		if (shader->inst[3] & NV40_VP_INST_VEC_WRITEMASK_X) mask |= SMASK_X;
		if (shader->inst[3] & NV40_VP_INST_VEC_WRITEMASK_Y) mask |= SMASK_Y;
		if (shader->inst[3] & NV40_VP_INST_VEC_WRITEMASK_Z) mask |= SMASK_Z;
		if (shader->inst[3] & NV40_VP_INST_VEC_WRITEMASK_W) mask |= SMASK_W;
	}

	return mask;
}

static unsigned int
NV40VPGetSourceHW(ShaderFunc *shader, int merged, int pos)
{
	struct _op_xlat *opr;
	unsigned int src;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr)
		return -1;

	switch (opr->srcpos[pos]) {
	case 0:
		src = ((shader->inst[1] & NV40_VP_INST_SRC0H_MASK)
				>> NV40_VP_INST_SRC0H_SHIFT)
			<< NV40_VP_SRC0_HIGH_SHIFT;
		src |= ((shader->inst[2] & NV40_VP_INST_SRC0L_MASK)
				>> NV40_VP_INST_SRC0L_SHIFT);
		break;
	case 1:
		src = ((shader->inst[2] & NV40_VP_INST_SRC1_MASK)
				>> NV40_VP_INST_SRC1_SHIFT);
		break;
	case 2:
		src = ((shader->inst[2] & NV40_VP_INST_SRC2H_MASK)
				>> NV40_VP_INST_SRC2H_SHIFT)
			<< NV40_VP_SRC2_HIGH_SHIFT;
		src |= ((shader->inst[3] & NV40_VP_INST_SRC2L_MASK)
				>> NV40_VP_INST_SRC2L_SHIFT);
		break;
	default:
		src = -1;
	}

	return src;
}

static ShaderRegFile
NV40VPGetSourceFile(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	struct _op_xlat *opr;
	int file;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr || opr->srcpos[pos] == -1)
		return -1;

	switch (opr->srcpos[pos]) {
	case SPOS_TEXUNIT: return SRF_TEXUNIT;
	case SPOS_ADDRESS: return SRF_ADDRESS;
	default:
		src  = shader->GetSourceHW(shader, merged, pos);
		file = (src & NV40_VP_SRC_REG_TYPE_MASK) >> NV40_VP_SRC_REG_TYPE_SHIFT;

		switch (file) {
		case NV40_VP_SRC_REG_TYPE_TEMP : return SRF_TEMP;
		case NV40_VP_SRC_REG_TYPE_INPUT: return SRF_ATTRIB;
		case NV40_VP_SRC_REG_TYPE_CONST: return SRF_CONST;
		default:
			return SRF_UNKNOWN;
		}
	}
}

static int
NV40VPGetSourceID(ShaderFunc *shader, int merged, int pos)
{
	switch (shader->GetSourceFile(shader, merged, pos)) {
	case SRF_ATTRIB:
		switch ((shader->inst[1] & NV40_VP_INST_INPUT_SRC_MASK)
				>> NV40_VP_INST_INPUT_SRC_SHIFT) {
		case NV40_VP_INST_IN_POS   : return SR_POSITION;
		case NV40_VP_INST_IN_WEIGHT: return SR_WEIGHT;
		case NV40_VP_INST_IN_NORMAL: return SR_NORMAL;
		case NV40_VP_INST_IN_COL0  : return SR_COL0;
		case NV40_VP_INST_IN_COL1  : return SR_COL1;
		case NV40_VP_INST_IN_FOGC  : return SR_FOGCOORD;
		case NV40_VP_INST_IN_TC(0) : return SR_TEXCOORD0;
		case NV40_VP_INST_IN_TC(1) : return SR_TEXCOORD1;
		case NV40_VP_INST_IN_TC(2) : return SR_TEXCOORD2;
		case NV40_VP_INST_IN_TC(3) : return SR_TEXCOORD3;
		case NV40_VP_INST_IN_TC(4) : return SR_TEXCOORD4;
		case NV40_VP_INST_IN_TC(5) : return SR_TEXCOORD5;
		case NV40_VP_INST_IN_TC(6) : return SR_TEXCOORD6;
		case NV40_VP_INST_IN_TC(7) : return SR_TEXCOORD7;
		default: return -1;
		}
		break;
	case SRF_CONST:
		return ((shader->inst[1] & NV40_VP_INST_CONST_SRC_MASK)
				>> NV40_VP_INST_CONST_SRC_SHIFT);
	case SRF_TEMP:
	{
		unsigned int src;

		src = shader->GetSourceHW(shader, merged, pos);
		return ((src & NV40_VP_SRC_TEMP_SRC_MASK) >> NV40_VP_SRC_TEMP_SRC_SHIFT);
	}
	default:
		return -1;
	}
}

static int
NV40VPGetSourceNegate(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;

	src = shader->GetSourceHW(shader, merged, pos);

	if (src == -1)
		return -1;
	return ((src & NV40_VP_SRC_NEGATE) ? 1 : 0);
}

static ShaderRegSwz
NV40VPGetSourceSwizzle(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	int swz;

	src = shader->GetSourceHW(shader, merged, pos);
	swz = (src & NV40_VP_SRC_SWZ_ALL_MASK) >> NV40_VP_SRC_SWZ_ALL_SHIFT;
	return NV20VPTXSwizzle(swz);
}

static int
NV40VPGetSourceIndexed(ShaderFunc *shader, int merged, int pos)
{
	switch (shader->GetSourceFile(shader, merged, pos)) {
	case SRF_ATTRIB:
		return ((shader->inst[0] & NV40_VP_INST_INDEX_INPUT) ? 1 : 0);
	case SRF_CONST:
		return ((shader->inst[3] & NV40_VP_INST_INDEX_CONST) ? 1 : 0);
	default:
		return 0;
	}
}

static ShaderSwzComp
NV40VPGetAddressRegSwizzle(ShaderFunc *shader)
{
	ShaderSwzComp swz;

	swz = NV20VP_TX_SWIZZLE[(shader->inst[0] & NV40_VP_INST_ADDR_SWZ_MASK)
		>> NV40_VP_INST_ADDR_SWZ_SHIFT];
	return swz;
}

static int
NV40VPSupportsConditional(ShaderFunc *shader)
{
	/*FIXME: Is this true of all ops? */
	return 1;
}

static int
NV40VPGetConditionUpdate(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV40_VP_INST_COND_UPDATE_ENABLE) ? 1 : 0);
}

static int
NV40VPGetConditionTest(ShaderFunc *shader)
{
	int op;

	/* The condition test is unconditionally enabled on some
	 * instructions. ie: the condition test bit does *NOT* have
	 * to be set.
	 *
	 * FIXME: check other relevant ops for this situation.
	 */
	op = shader->GetOpcodeHW(shader, 1);
	switch (op) {
	case NV40_VP_INST_OP_BRA:
		return 1;
	default:
		return ((shader->inst[0] & NV40_VP_INST_COND_TEST_ENABLE) ? 1 : 0);
	}
}

static ShaderCond
NV40VPGetCondition(ShaderFunc *shader)
{
	int cond;

	cond = ((shader->inst[0] & NV40_VP_INST_COND_MASK)
			>> NV40_VP_INST_COND_SHIFT);

	switch (cond) {
	case NV40_VP_INST_COND_FL: return SC_FL;
	case NV40_VP_INST_COND_LT: return SC_LT;
	case NV40_VP_INST_COND_EQ: return SC_EQ;
	case NV40_VP_INST_COND_LE: return SC_LE;
	case NV40_VP_INST_COND_GT: return SC_GT;
	case NV40_VP_INST_COND_NE: return SC_NE;
	case NV40_VP_INST_COND_GE: return SC_GE;
	case NV40_VP_INST_COND_TR: return SC_TR;
	default:
		return SC_UNKNOWN;
	}
}

static ShaderRegSwz
NV40VPGetCondRegSwizzle(ShaderFunc *shader)
{
	int swz;

	swz = (shader->inst[0] & NV40_VP_INST_COND_SWZ_ALL_MASK)
		>> NV40_VP_INST_COND_SWZ_ALL_SHIFT;
	return NV20VPTXSwizzle(swz);
}

static int
NV40VPGetCondRegID(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV40_VP_INST_COND_REG_SELECT_1) ? 1 : 0);
}

static int
NV40VPGetBranch(ShaderFunc *shader)
{
	int addr;

	addr = ((shader->inst[2] & NV40_VP_INST_IADDRH_MASK)
			>> NV40_VP_INST_IADDRH_SHIFT) << 3;
	addr |= ((shader->inst[3] & NV40_VP_INST_IADDRL_MASK)
			>> NV40_VP_INST_IADDRL_SHIFT);
	return addr;
}

void
NV40VPInitShaderFuncs(ShaderFunc *shader)
{
	/* Inherit NV30 VP code, we share some of it */
	NV30VPInitShaderFuncs(shader);

	/* Add extra opcodes for NV40+ */
//	MOD_OPCODE(NVVP_TX_VOP, NV40_VP_INST_OP_TXWHAT, SOP_TEX  ,  0,  4, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV40_VP_INST_OP_PUSHA , SOP_PUSHA,  3, -1, -1);
	MOD_OPCODE(NVVP_TX_SOP, NV40_VP_INST_OP_POPA  , SOP_POPA , -1, -1, -1);

	shader->HasMergedInst			= NV40VPHasMergedInst;
	shader->GetOpcodeHW				= NV40VPGetOpcodeHW;

	shader->GetDestFile				= NV40VPGetDestFile;
	shader->GetDestID				= NV40VPGetDestID;
	shader->GetDestMask				= NV40VPGetDestMask;

	shader->GetSourceHW				= NV40VPGetSourceHW;
	shader->GetSourceFile			= NV40VPGetSourceFile;
	shader->GetSourceID				= NV40VPGetSourceID;
	shader->GetSourceNegate			= NV40VPGetSourceNegate;
	shader->GetSourceSwizzle		= NV40VPGetSourceSwizzle;
	shader->GetSourceIndexed		= NV40VPGetSourceIndexed;

	shader->GetRelAddressSwizzle	= NV40VPGetAddressRegSwizzle;

	shader->SupportsConditional		= NV40VPSupportsConditional;
	shader->GetConditionUpdate		= NV40VPGetConditionUpdate;
	shader->GetConditionTest		= NV40VPGetConditionTest;
	shader->GetCondition			= NV40VPGetCondition;
	shader->GetCondRegSwizzle		= NV40VPGetCondRegSwizzle;
	shader->GetCondRegID			= NV40VPGetCondRegID;

	shader->GetBranch				= NV40VPGetBranch;
}

