#ifndef __SHADER_COMMON_H__
#define __SHADER_COMMON_H__

typedef struct _ShaderFunc ShaderFunc;

typedef enum {
	SRF_NONE,
	SRF_TEMP,
	SRF_ATTRIB,
	SRF_CONST,
	SRF_RESULT,
	SRF_ADDRESS,
	SRF_TEXUNIT,
	SRF_UNKNOWN
} ShaderRegFile;

typedef enum {
	SOP_UNKNOWN = 0,
	SOP_NOP,
	SOP_ABS, SOP_ADD, SOP_ARA, SOP_ARL, SOP_ARR,
	SOP_BRA, SOP_BRK,
	SOP_CAL, SOP_CMP, SOP_COS,
	SOP_DDX, SOP_DDY, SOP_DIV, SOP_DP2, SOP_DP2A, SOP_DP3, SOP_DP4, SOP_DPH,
	SOP_DST,
	SOP_EX2, SOP_EXP,
	SOP_FLR, SOP_FRC,
	SOP_IF,
	SOP_KIL,
	SOP_LG2, SOP_LIT, SOP_LOG, SOP_LOOP, SOP_LRP,
	SOP_MAD, SOP_MAX, SOP_MIN, SOP_MOV, SOP_MUL,
	SOP_NRM,
	SOP_PK2H, SOP_PK2US, SOP_PK4B, SOP_PK4UB, SOP_POW, SOP_POPA, SOP_PUSHA,
	SOP_RCC, SOP_RCP, SOP_REP, SOP_RET, SOP_RFL, SOP_RSQ,
	SOP_SCS, SOP_SEQ, SOP_SFL, SOP_SGE, SOP_SGT, SOP_SIN, SOP_SLE, SOP_SLT,
	SOP_SNE, SOP_SSG, SOP_STR, SOP_SUB, SOP_SWZ,
	SOP_TEX, SOP_TXB, SOP_TXD, SOP_TXL, SOP_TXP,
	SOP_UP2H, SOP_UP2US, SOP_UP4B, SOP_UP4UB,
	SOP_X2D, SOP_XPD
} ShaderOpcode;

typedef enum {
	SP_FLOAT32,
	SP_FLOAT16,
	SP_FIXED12,
	SP_UNKNOWN
} ShaderPrecision;

typedef enum {
	SSC_X, SSC_Y, SSC_Z, SSC_W, 
} ShaderSwzComp;

typedef struct _ShaderRegSwz {
	ShaderSwzComp x, y, z, w;
} ShaderRegSwz;

typedef enum {
	SR_POSITION,
	SR_WEIGHT,
	SR_NORMAL,
	SR_COL0,
	SR_COL1,
	SR_BFC0,
	SR_BFC1,
	SR_FOGCOORD,
	SR_POINTSZ,
	SR_TEXCOORD0,
	SR_TEXCOORD1,
	SR_TEXCOORD2,
	SR_TEXCOORD3,
	SR_TEXCOORD4,
	SR_TEXCOORD5,
	SR_TEXCOORD6,
	SR_TEXCOORD7,
	SR_FRAGDATA0,
	SR_FRAGDATA1,
	SR_FRAGDATA2,
	SR_FRAGDATA3,
	SR_CLIP0,
	SR_CLIP1,
	SR_CLIP2,
	SR_CLIP3,
	SR_CLIP4,
	SR_CLIP5,
	SR_CLIP6,
	SR_FACING,
	SR_UNKNOWN
} ShaderFixedReg;

typedef enum {
	SC_FL, SC_LT, SC_EQ, SC_LE, SC_GT, SC_NE, SC_GE, SC_TR,
	SC_UNKNOWN
} ShaderCond;

#define SMASK_X (1<<0)
#define SMASK_Y (1<<1)
#define SMASK_Z (1<<2)
#define SMASK_W (1<<3)
#define SMASK_ALL (SMASK_X|SMASK_Y|SMASK_Z|SMASK_W)

#define SPOS_ADDRESS 3
#define SPOS_TEXUNIT 4
struct _op_xlat {
	unsigned int	NV;
	ShaderOpcode	SOP;
	int			srcpos[3];
};
#define MOD_OPCODE(t,hw,sop,s0,s1,s2) do { \
	t[hw].NV = hw; \
	t[hw].SOP = sop; \
	t[hw].srcpos[0] = s0; \
	t[hw].srcpos[1] = s1; \
	t[hw].srcpos[2] = s2; \
} while(0)

extern unsigned int NVVP_TX_VOP_COUNT;
extern unsigned int NVVP_TX_SOP_COUNT;
extern struct _op_xlat NVVP_TX_VOP[];
extern struct _op_xlat NVVP_TX_SOP[];

extern unsigned int NVFP_TX_AOP_COUNT;
extern unsigned int NVFP_TX_BOP_COUNT;
extern struct _op_xlat NVFP_TX_AOP[];
extern struct _op_xlat NVFP_TX_BOP[];

struct _ShaderFunc {
	unsigned int *	inst;
	struct _op_xlat*(*GetOPTXRec)		(ShaderFunc *, int merged);

	int		(*HasMergedInst)	(ShaderFunc *);
	int		(*IsLastInst)		(ShaderFunc *);
	int		(*GetOffsetNext)	(ShaderFunc *);

	int		(*GetOpcodeSlot)	(ShaderFunc *, int merged);
	unsigned int	(*GetOpcodeHW)		(ShaderFunc *, int slot);
	ShaderOpcode	(*GetOpcode)		(ShaderFunc *, int merged);

	ShaderPrecision	(*GetPrecision)		(ShaderFunc *);
	int		(*GetSaturate)		(ShaderFunc *);

	ShaderRegFile	(*GetDestFile)		(ShaderFunc *, int merged);
	unsigned int	(*GetDestID)		(ShaderFunc *, int merged);
	unsigned int	(*GetDestMask)		(ShaderFunc *, int merged);

	unsigned int	(*GetSourceHW)		(ShaderFunc *, int merged, int pos);
	ShaderRegFile	(*GetSourceFile)	(ShaderFunc *, int merged, int pos);
	int		(*GetSourceID)		(ShaderFunc *, int merged, int pos);
	int		(*GetSourceNegate)	(ShaderFunc *, int merged, int pos);
	int		(*GetSourceAbs)		(ShaderFunc *, int merged, int pos);
	ShaderRegSwz	(*GetSourceSwizzle)	(ShaderFunc *, int merged, int pos);
	int		(*GetSourceIndexed)	(ShaderFunc *, int merged, int pos);
	void		(*GetSourceConstVal)	(ShaderFunc *, int merged, int pos, float *val);
	int		(*GetSourceScale)	(ShaderFunc *, int merged, int pos);

	int		(*GetRelAddressRegID)	(ShaderFunc *);
	ShaderSwzComp	(*GetRelAddressSwizzle)	(ShaderFunc *);

	int		(*SupportsConditional)	(ShaderFunc *);
	int		(*GetConditionUpdate)	(ShaderFunc *);
	int		(*GetConditionTest)	(ShaderFunc *);
	ShaderCond	(*GetCondition)		(ShaderFunc *);
	ShaderRegSwz	(*GetCondRegSwizzle)	(ShaderFunc *);
	int		(*GetCondRegID)		(ShaderFunc *);

	int		(*GetBranch)		(ShaderFunc *);
	int		(*GetBranchElse)	(ShaderFunc *);
	int		(*GetBranchEnd)		(ShaderFunc *);

	int		(*GetLoopCount)		(ShaderFunc *);
	int		(*GetLoopInitial)	(ShaderFunc *);
	int		(*GetLoopIncrement)	(ShaderFunc *);
};

extern void NV20VPInitShaderFuncs(ShaderFunc *);
extern void NV30VPInitShaderFuncs(ShaderFunc *);
extern void NV40VPInitShaderFuncs(ShaderFunc *);

extern void NV30FPInitShaderFuncs(ShaderFunc *);
extern void NV40FPInitShaderFuncs(ShaderFunc *);

#endif
